from .ninjax import __version__

# Basics
from .ninjax import Module
from .ninjax import Variable
from .ninjax import pure
from .ninjax import init
from .ninjax import seed

# Transforms
from .ninjax import grad
from .ninjax import cond
from .ninjax import scan
from .ninjax import checkpoint

# Advanced
from .ninjax import context
from .ninjax import creating
from .ninjax import scope

# Integrations
from .ninjax import FromHaiku
from .ninjax import FromFlax
from .ninjax import FromOptax



import jax
import jax.numpy as jnp
import ninjax as nj
import pytest


class TestAttrs:

  def test_attrs(self):
    class MyModule(nj.Module):
      foo: int
      bar: float = 12.5
      baz: str = None
      other = True
    module = MyModule(foo=42, baz='baz', name='module')
    assert module.foo == 42
    assert module.bar == 12.5
    assert module.baz == 'baz'

  def test_class_variable(self):
    class MyModule(nj.Module):
      foo: int
      other = True
    module = MyModule(foo=42, name='module')
    assert module.foo == 42
    assert module.other == True
    with pytest.raises(TypeError):
      MyModule(other=False, name='module')

  def test_keyword_only(self):
    class MyModule(nj.Module):
      foo: int
    with pytest.raises(TypeError):
      module = MyModule(42, name='module')

  def test_with_default(self):
    class MyModule(nj.Module):
      foo: int = 42
    module = MyModule(name='module')
    assert module.foo == 42
    module = MyModule(foo=12, name='module')
    assert module.foo == 12

  def test_without_default(self):
    class MyModule(nj.Module):
      foo: int
    module = MyModule(foo=12, name='module')
    assert module.foo == 12
    with pytest.raises(TypeError):
      MyModule(name='module')

  def test_immutable(self):
    class MyModule(nj.Module):
      foo: int
    module = MyModule(foo=42, name='module')
    assert module.foo == 42
    with pytest.raises(AttributeError):
      module.foo = 12

  def test_type_check(self):
    class MyModule(nj.Module):
      foo: int
    with pytest.raises(TypeError):
      module = MyModule(foo=1.5, name='module')

  def test_with_init(self):
    class MyModule(nj.Module):
      foo: int
      def __init__(self, bar):
        self.bar = bar
    module = MyModule(12, foo=42, name='module')
    assert module.foo == 42
    assert module.bar == 12
    module = MyModule(foo=42, bar=12, name='module')
    assert module.foo == 42
    assert module.bar == 12



import jax
import jax.numpy as jnp
import ninjax as nj
import pytest


class TestInit:

  def test_empty(self):
    fun = nj.pure(lambda: jnp.array(42))
    state = nj.init(fun)({})
    assert state == {}
    assert jax.jit(fun)(state)[1] == 42

  def test_variable(self):
    v = nj.Variable(jnp.ones, (), jnp.int32, name='v')
    read = nj.pure(v.read)
    write = nj.pure(v.write)
    state = nj.init(read)({})
    assert state == {'v/value': 1}
    assert read(state)[1] == 1
    state, _ = write(state, 42)
    assert state == {'v/value': 42}
    assert read(state)[1] == 42

  def test_random(self):
    @nj.pure
    def program():
      v = nj.Variable(jax.random.uniform, nj.seed(), name='v')
      return v.read()
    with pytest.raises(ValueError):
      nj.init(program)({})
    state = nj.init(program)({}, seed=0)
    assert set(state.keys()) == {'v/value'}
    assert jnp.round(state['v/value'], 4) == 0.9892
    state, value = jax.jit(program)(state, seed=1)
    assert jnp.round(value, 4) == 0.9892

  def test_static(self):
    traces = []
    @nj.pure
    def program(mode='a'):
      traces.append(mode)
      a = nj.Variable(jnp.array, 42, name='a')
      b = nj.Variable(jnp.array, 21, name='b')
      if mode == 'a':
        return a.read()
      else:
        return b.read()
    state = nj.init(program, static_argnames=['mode'])({}, mode='a')
    assert traces == ['a']
    state = nj.init(program, static_argnames=['mode'])(state, mode='b')
    assert traces == ['a', 'b']
    assert state == {'a/value': 42, 'b/value': 21}
    apply = jax.jit(program, static_argnames=['mode'])
    assert apply(state, mode='a')[1] == 42
    assert traces == ['a', 'b', 'a']
    assert apply(state, mode='a')[1] == 42
    assert traces == ['a', 'b', 'a']
    assert apply(state, mode='b')[1] == 21
    assert traces == ['a', 'b', 'a', 'b']
    assert apply(state, mode='b')[1] == 21
    assert traces == ['a', 'b', 'a', 'b']

  def test_overlap(self):
    a = nj.Variable(jnp.zeros, (), jnp.int32, name='a')
    b = nj.Variable(jnp.zeros, (), jnp.int32, name='b')
    c = nj.Variable(jnp.zeros, (), jnp.int32, name='c')
    @nj.pure
    def foo():
      a.write(a.read() + 1)
      return b.read()
    @nj.pure
    def bar():
      c.write(a.read() + 1)
      return a.read()
    state = {}
    state = nj.init(foo)(state)
    state = nj.init(bar)(state)
    assert state == {'a/value': 0, 'b/value': 0, 'c/value': 1}



import jax
import jax.numpy as jnp
import ninjax as nj


class TestJit:

  def test_constants(self):
    fun = jax.jit(nj.pure(lambda: jnp.array(42)))
    assert not hasattr(fun, 'keys')
    assert fun({})[1] == 42
    # assert fun.keys == set()
    assert fun({})[1] == 42
    # assert fun.keys == set()

  def test_variables(self):
    v = nj.Variable(jnp.ones, (), jnp.int32, name='v')
    state = nj.init(v.read)({})
    read = jax.jit(nj.pure(v.read))
    write = jax.jit(nj.pure(v.write))
    # assert not hasattr(read, 'keys')
    # assert not hasattr(write, 'keys')
    state = read(state)[0]
    # assert read.keys == {'v/value'}
    assert state == {'v/value': 1}
    state = write(state, 42)[0]
    # assert write.keys == {'v/value'}
    assert state == {'v/value': 42}
    value = read(state)[1]
    assert value == 42
    assert state == {'v/value': 42}

  def test_statics(self):
    @nj.pure
    def program(x, mode='train'):
      return x if mode == 'train' else 2 * x
    fun = jax.jit(program, static_argnames=['mode'])
    assert fun({}, jnp.array(1), mode='train')[1] == 1
    assert fun({}, jnp.array(1), mode='train')[1] == 1
    assert fun({}, jnp.array(1), mode='eval')[1] == 2
    assert fun({}, jnp.array(1), mode='eval')[1] == 2

  def test_overlap(self):
    v1 = nj.Variable(jnp.zeros, (), jnp.int32, name='v1')
    v2 = nj.Variable(jnp.zeros, (), jnp.int32, name='v2')
    v3 = nj.Variable(jnp.zeros, (), jnp.int32, name='v3')
    def foo():
      v1.write(v1.read() + 1)
      return v2.read()
    def bar():
      v3.write(v1.read() + 1)
      return v1.read()
    state = {}
    state = nj.init(foo)(state)
    state = nj.init(bar)(state)
    assert state == {'v1/value': 0, 'v2/value': 0, 'v3/value': 1}
    foo = jax.jit(nj.pure(foo))
    bar = jax.jit(nj.pure(bar))
    state = foo(state)[0]
    assert state == {'v1/value': 1, 'v2/value': 0, 'v3/value': 1}
    state = bar(state)[0]
    assert state == {'v1/value': 1, 'v2/value': 0, 'v3/value': 2}
    state = foo(state)[0]
    assert state == {'v1/value': 2, 'v2/value': 0, 'v3/value': 2}
    state = bar(state)[0]
    assert state == {'v1/value': 2, 'v2/value': 0, 'v3/value': 3}

  def test_side_effect(self):
    counter = nj.Variable(jnp.array, 0, name='counter')
    @nj.pure
    def program(x):
      counter.write(counter.read() + 1)
      return counter.read()
    state = nj.init(program)({}, 0)
    fun = jax.jit(program)
    for index in range(1, 4):
      state, value = fun(state, jnp.array(2.0))
      assert value == index



import jax.numpy as jnp
import ninjax as nj


class TestCond:

  def test_basic(self):
    @nj.pure
    def program(x):
      return nj.cond(x, lambda: 12, lambda: 42)
    assert program({}, True)[1] == 12
    assert program({}, False)[1] == 42

  def test_state(self):
    v = nj.Variable(jnp.ones, (), jnp.int32, name='v')
    w = nj.Variable(jnp.ones, (), jnp.int32, name='w')
    @nj.pure
    def program(x):
      def true_fn():
        return v.read() + w.read()
      def false_fn():
        v.write(42)
        return 12
      return nj.cond(x, true_fn, false_fn)
    state = nj.init(program)({}, True)
    assert state == {'v/value': 1, 'w/value': 1}
    state, out = program(state, True)
    assert out == 2
    state, out = program(state, False)
    assert out == 12
    state, out = program(state, True)
    assert out == 43

  def test_seeds(self):
    @nj.pure
    def program(x):
      return nj.cond(x, nj.seed, nj.seed)
    _, draw1 = program({}, True, seed=0)
    _, draw2 = program({}, False, seed=0)
    assert jnp.array(draw1).shape == (2,)
    assert jnp.array(draw2).shape == (2,)
    assert (draw1 != draw2).all()



import jax.numpy as jnp
import ninjax as nj


class TestScan:

  def test_constants(self):
    def program():
      def body(carry, x):
        carry = carry + x
        return carry, carry
      return nj.scan(body, 0, jnp.array([1, 2, 3, 4, 5]))
    _, (carry, ys) = nj.pure(program)({})
    assert carry == 15
    assert list(ys) == [1, 3, 6, 10, 15]

  def test_read_state(self):
    v = nj.Variable(jnp.zeros, (), jnp.int32, name='v')
    def program():
      def body(carry, x):
        y = x + v.read()
        return carry + 1, y
      return nj.scan(body, 0, jnp.array([1, 1, 1]))
    _, (carry, ys) = nj.pure(program)({}, create=True)
    assert carry == 3
    assert list(ys) == [1, 1, 1]

  def test_write_state(self):
    v = nj.Variable(jnp.zeros, (), jnp.int32, name='v')
    def program():
      def body(carry, x):
        y = x + v.read()
        v.write(y)
        return carry + 1, y
      return nj.scan(body, 0, jnp.array([1, 1, 1]))
    _, (carry, ys) = nj.pure(program)({}, create=True)
    assert carry == 3
    assert list(ys) == [1, 2, 3]

  def test_seed(self):
    def program():
      def body(carry, x):
        return carry, nj.seed()
      return nj.scan(body, 0, jnp.array([1, 1, 1]))[1]
    _, keys = nj.pure(program)({}, seed=0, create=True)
    assert jnp.array(keys).shape == (3, 2)
    assert (keys[0] != keys[1]).all()
    assert (keys[0] != keys[2]).all()
    assert (keys[1] != keys[2]).all()
